

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <cmath>
#include "plink.h"
#include "options.h"

using namespace std;



bool Plink::isAffPair(Individual *p1, Individual *p2)
{
  
  // Assume 1=unaffected
  //        2=affected 
  
  // only not skip if both individuals are affected:
  
  if (p1->phenotype == 2 && p2->phenotype == 2) return true;
  else return false;
  
}

bool Plink::isUnAffPair(Individual *p1, Individual *p2)
{
  
  // Assume 1=unaffected
  //        2=affected 
  
  // only not skip if both individuals are affected:
  
  if (p1->phenotype == 1 && p2->phenotype == 1) return true;
  else return false;
  
}


void Plink::calcAffPair(vector<vector<double> > & pihat)
{ 
  // pihat[pair][position]
  
  int npos = pihat[0].size();
  
  int R = par::replicates;
 
  // Store each regression r^2 in here;
  // original      : res[0][position]
  // permutation p : res[p][position]

  vector<double> pv(npos);
  vector<double> pv2(npos); // 2-sided p-values

  vector<double> maxres;
  vector<double> res;
  
  // Precalculate mean and variance of pihat and phenotype
  preCalcRegression_PHENO(phenotype);
  
  if(par::verbose)
    {  
      for (int i=0;i<v_pihat.size();i++)
	cout << i << " V pihat = " << v_pihat[i] << "\n";
      
      // Display entire dataset
      for (int i=0;i<pihat.size();i++)
	{
	  cout << "PIHAT " << i << "\t";
	  for (int j=0;j<pihat[i].size();j++)
	    cout << pihat[i][j] << " ";
	  cout << "\n";
	}
      
      cout << "\n";
      for (int i=0;i<phenotype.size();i++)
	cout << "PHENO " << i << "\t" << phenotype[i] << "\n";
      cout << "\n";
    }
  
  // Save original data  
  res = doRegression(npos,phenotype);
  
  // Save original results
  original.push_back(res);
  
  // Permuted data
  for (int p=1; p<=R; p++)
    {
      cerr << "Regression permutation: "
	   << p
	   << " of " 
	   << R 
	   << "             \r";
      
      
      // Permute phenotype pairs
      vector<long int> i(phenotype.size());
      vector<double> perm(phenotype.size());
      permute(i);
      for (int j=0; j<phenotype.size(); j++)
	perm[j] = phenotype[i[j]];
      
      // Perform analysis at each position
      double mx=0;

      vector<double> pres = doRegression(npos,perm);
	  
      for (int l=0; l<npos; l++)
	{
	  // Does this exceed the original?
	  if (pres[l] >= res[l]) pv[l]++;
	  
	  // tmp : 2-sided test
	  if (abs(pres[l]) >= abs(res[l])) pv2[l]++;
	  
	  // Is this the maximum?
	  if (pres[l]>mx) mx=pres[l];
	  
	}
      
      // Save maximum statistic
      maxres.push_back(mx);
      
    }
  
  // Save max-values for genome-wide comparison
  for (int p=0; p<R; p++)
    if (maxres[p] >= maxr2[p]) maxr2[p]=maxres[p];
  

  cerr << "\n";

  // Maximum-adjusted p-values
  for (int l=0; l<npos; l++)
    {

      int maxpv=1;
      
      // Empircally-corrected p-value
      for (int p=0; p<R; p++)
	if (maxres[p] >= res[l]) maxpv++;

      double p1, p2;
      string n1, n2;
      
      if (m1[l]==-1) 
	{
	  p1 = locus[par::run_start]->pos - par::fringe;
	  n1 = "fringe";
	}
      else 
	{
	  p1 = locus[m1[l]]->pos;
	  n1 = locus[m1[l]]->name;
	}

      if (m2[l]==-1) 
	{
	  p2 = locus[par::run_end]->pos + par::fringe;
	  n2 = "fringe";
	}
      else 
	{
	  p2 = locus[m2[l]]->pos;
	  n2 = locus[m2[l]]->name;
	}

      if (m1[l]==-1 && m2[l]==-1)
	{
	  p1 = p2 = 0;
	  n1 = "Genomewide";
	  n2 = "IBD";
	}
      
      double d1 = p1 + pos[l] * (p2-p1);
	    
      if (res[l] != res[l]) 
	cout << "R " 
	     << par::run_chr << " " 
	     << n1 << " " 
	     << n2 << " " 
	     << d1 << "  "
	     << "NaN NaN NaN\n";
      else
	cout << "R " 
	     << par::run_chr << " " 
	     << n1 << " " 
	     << n2 << " " 
	     << d1 << "  "
	     << res[l] << " " 
	     << (double)(pv[l]+1)/(double)(R+1) << " "
	     << (double)(pv2[l]+1)/(double)(R+1) << " "
	     << (double)maxpv/(double)(R+1) << "\n";
	      
    }



  
  // L = lambda, sibling relative risk... assume=1
  // Z = X*X + Y*Y - ( (1-2*K+K*L ) / ( K * L * (1-K) )  )*X*Y
  
  //regress Z on pihat

}



void Plink::preCalcAffPair()
{
  
  /////////////////////////////////////
  // Precalculate means and variances of IBD only
  // and standardize

  // Number of pairs
  int n = pihat.size();

  // Number of positions
  int l = pihat[0].size();
  
  m_pihat.resize(0);
  v_pihat.resize(0);
  
  // For each locus
  for (int j=0; j<l; j++)
    {
      double m_p=0;
      double v_p=0;
      
      // Pihat mean
      for (int i=0; i<n; i++) 
	m_p += pihat[i][j]; 
      m_p /= (double)n;
      m_pihat.push_back(m_p); 
      
      // Pihat variance
      for (int i=0; i<n; i++)
	v_p += (pihat[i][j] - m_p) * ( pihat[i][j] - m_p);
      v_p /= (double)n;
      v_pihat.push_back(v_p);

      // Standardize
      for (int i=0; i<n; i++)
	pihat[i][j] =  ( pihat[i][j] - m_p ) / sqrt(v_p);
      
    }  

}

vector<double> Plink::doAffPair(int npos)
{

  vector<double> rv;

  // Number of pairs
  int n = pihat.size();


  // calculate some statistic based on deviation from expected pihat
  // challenge: no room for permutation now?

  for (int l=0; l<npos; l++)
    {
     // pihat[i][l]
      
      //      rv.push_back( {statistic} );
    }
  
  return rv;
  
}
